
/*  j2anim.c    Version 1.0, by Jack Boyce (jboyce@tybalt.caltech.edu)    */
/*                                                                        */
/*  This program converts juggling patterns from the notation used by the */
/*  pattern-generating program J2 into formats readable by JUG and        */
/*  JUGGLEPRO.                                                            */


#define NUM_BALLS   10

#define RIGHT        0
#define LEFT         1

#define ASYNCH       0         /* different input modes */
#define SYNCH        1
#define PASSING      2
#define CUSTOM       3

#define JUG          0         /* destination formats */
#define JUGGLEPRO    1

#define SYNTAX       0         /* different error conditions */
#define MEMORY       1
#define TOO_MANY     2
#define NO_JUG       3
#define NO_JP        4

char *colors[NUM_BALLS] = { "red", "green", "blue", "yellow", "violet",
                            "cyan", "white", "black", "brown", "orange" };

struct throw {
   int value;
   int to;
};

struct ball {
   int ticks;
   int dest_hand;
   int loop_ticks;
   int loop_dest;
};

struct snapshot {       /* structure for JUG multiplexing */
   struct snapshot *prev;
   int pos;
   int time;
   int *ticks;
   int *dest;
};

int n = 0, l = 0, mode, hands, max_occupancy = 1;

void die();


/*  The following routine parses the input pattern the first time.  */
/*  It finds n, l, and the max_occupancy.  With these the program   */
/*  can allocate memory space.                                      */

void parse1(pattern, dest_format)
char *pattern;
int dest_format;
{
   int i, j, in_brackets = 0, in_braces = 0, occupancy, got_divider;
   char ch;

   for (i = 0; ch = pattern[i]; i++) {
      if (!in_brackets && !in_braces)
         l++;
      switch (ch) {
         case '[':
            if (in_brackets || ((mode != ASYNCH) && (!in_braces)))
               die(SYNTAX);
            in_brackets = 1;
            occupancy = 0;
            break;
         case ']':
            if (!in_brackets || ((mode != ASYNCH) && (!in_braces)))
               die(SYNTAX);
            in_brackets = 0;
            if (occupancy > max_occupancy)
               max_occupancy = occupancy;
            break;
         case '<':
            if (in_braces || (mode != PASSING))
               die(SYNTAX);
            in_braces = 1;
            got_divider = 0;
            break;
         case '(':
            if (in_braces || (mode != SYNCH))
               die(SYNTAX);
            in_braces = 1;
            got_divider = 0;
            break;
         case '>':
            if (!in_braces || (mode != PASSING) || (!got_divider))
               die(SYNTAX);
            in_braces = 0;
            break;
         case ')':
            if (!in_braces || (mode != SYNCH) || (!got_divider))
               die(SYNTAX);
            in_braces = 0;
            break;
         case ',':
            if (got_divider)
               die(SYNTAX);
            got_divider = 1;
         case 'x':
            if ((mode != SYNCH) || (!in_braces))
               die(SYNTAX);
            break;
         case '|':
            if (got_divider)
               die(SYNTAX);
            got_divider = 1;
         case 'p':
            if ((mode != PASSING) || (!in_braces))
               die(SYNTAX);
            break;
         default:
            if ((ch >= '0') && (ch <= '9'))
               j = (int)(ch - '0');
            else if ((ch >= 'A') && (ch <= 'Z'))
               j = 10 + (int)(ch - 'A');
            else if ((ch >= 'a') && (ch <= 'z'))
               j = 10 + (int)(ch - 'a');
            else
               die(SYNTAX);

            if (in_brackets)
               occupancy++;
            if ((mode != ASYNCH) && (!in_braces))
               die(SYNTAX);
            n += j;      /* add throw value to running sum */
            break;
      }
   }
   if (in_brackets || in_braces)
      die(SYNTAX);

   if (mode == SYNCH)
      l *= 2;
   if (n % l)
      die(SYNTAX);
   n /= l;               /* site swap average rule */

   if ((dest_format == JUG) && (n > NUM_BALLS))
      die(TOO_MANY);
}


/*  This parses the input string a second time, putting the pattern  */
/*  into the internal memory arrays.                                 */

void parse2(pattern, throw)
char *pattern;
struct throw ***throw;
{
   int i, j, in_brackets = 0, in_braces = 0;
   int pos = 0, occupancy = 0, got_divider;
   char ch;

   for (i = 0; ch = pattern[i]; i++) {
      switch (ch) {
         case '[':
            in_brackets = 1;
            break;
         case ']':
            in_brackets = 0;
            occupancy = 0;
            break;
         case '<':
         case '(':
            in_braces = 1;
            got_divider = 0;
            break;
         case '>':
         case ')':
            in_braces = 0;
            break;
         case ',':
         case '|':
            got_divider = 1;
            break;
         case 'x':
         case 'p':
            die(SYNTAX);
            break;
         default:
            if ((ch >= '0') && (ch <= '9'))
               j = (int)(ch - '0');
            else if ((ch >= 'A') && (ch <= 'Z'))
               j = 10 + (int)(ch - 'A');
            else if ((ch >= 'a') && (ch <= 'z'))
               j = 10 + (int)(ch - 'a');

            if (mode == ASYNCH) {
               throw[pos][0][occupancy].value = j;
               throw[pos][0][occupancy].to = 0;
            } else if (mode == SYNCH) {
               throw[pos][got_divider][occupancy].value = j;
               if (pattern[i + 1] == 'x') {
                  throw[pos][got_divider][occupancy].to = (got_divider?0:1);
                  i++;
               } else
                  throw[pos][got_divider][occupancy].to = got_divider;
            } else if (mode == PASSING) {
               throw[pos][got_divider][occupancy].value = j;
               if (pattern[i + 1] == 'p') {
                  throw[pos][got_divider][occupancy].to = (got_divider?0:1);
                  i++;
               } else
                  throw[pos][got_divider][occupancy].to = got_divider;
            }
            
            if (in_brackets)
               occupancy++;
            break;
      }
      if (!in_brackets && !in_braces)
         pos += ((mode == SYNCH) ? 2 : 1);

   }

}


/*  The next routines print out the pattern in memory in a format  */
/*  readable by S.Hawtin's JUG program.                            */

int do_time(print_time, time)     /* prints time labels, when needed */
int print_time, time;
{
   char buffer[20];

   if (print_time) {
      sprintf(buffer, "%d:", time);
      printf("%-4s ", buffer);
   } else
      printf("     ");

   return 0;     /* print_time is now 0 (don't print time any more) */
}


void print_jug(pattern, throw)    /* now the real routine */
char *pattern;
struct throw ***throw;
{
   int i, j, k, q, r, h = 0, balls_in_right = 0, throwing_hand = RIGHT;
   int depth = 0, time = 0, pos = 0, in_start = 1;
   int moveright, moveleft, loop_time, loop_pos = -1, print_time;
   struct ball *ball;
   struct snapshot *new_ptr, *snapshot_ptr = (struct snapshot *)0;

        /* First calculate the maximum throwing height in the pattern */

   for (i = 0; i < l; i++)
      for (j = 0; j < hands; j++)
         for (k = 0; (k < max_occupancy) && (q = throw[i][j][k].value); k++)
            if (q > h)
               h = q;

        /* Calculate the number of balls in right hand when we start */

   r = l - 1;
   for (i = 1; i <= h; i++) {
      for (j = 0; j < hands; j++) {
         for (k = 0; (k < max_occupancy) && (q = throw[r][j][k].value); k++) {
            if (q >= i) {
               if (mode == ASYNCH) {
                  if (!((q - i) % 2))
                     balls_in_right++;
               } else {                  /* synchronous juggling */
                  if (!throw[r][j][k].to)
                     balls_in_right++;
               }
            }
            if ((q - i) > depth)
               depth = q - i;
         }
      }
      r = (r ? (r - 1) : (l - 1));
   }
        /* Allocate memory space for ball variables */
                
   if ((ball = (struct ball *)malloc(n * sizeof(struct ball))) == 0)
      die(MEMORY);
   for (i = 0; i < n; i++) {
      ball[i].ticks = 0;    /* all balls are in hands */
      ball[i].dest_hand = ((i < balls_in_right) ? RIGHT : LEFT);
   }

   printf("!name site swap '%s'\n\n", pattern);
   printf("right");
   for (i = 0; i < n; i++)
      if (ball[i].dest_hand == RIGHT)
         printf(" %s", colors[i]);
   printf("\nleft");
   for (i = 0; i < n; i++)
      if (ball[i].dest_hand == LEFT)
         printf(" %s", colors[i]);
   printf("\n\n");

   
   while (1) {
      print_time = 1;
      
          /* check if we are moving out of startup phase */
	        
      if ((max_occupancy == 1) && (time == (2 * depth))) {
         in_start = 0;
	 loop_time = time;
	 loop_pos = pos;
         printf("\n");
      }
  
      if ((max_occupancy > 1) && ((time % 2) == 0)) {
         if ((new_ptr = (struct snapshot *)
                             malloc(sizeof(struct snapshot))) == 0)
            die(MEMORY);
         if ((new_ptr->ticks = (int *)malloc(n * sizeof(int))) == 0)
            die(MEMORY);
         if ((new_ptr->dest = (int *)malloc(n * sizeof(int))) == 0)
            die(MEMORY);
         new_ptr->prev = snapshot_ptr;
         snapshot_ptr = new_ptr;
         snapshot_ptr->pos = pos;
         snapshot_ptr->time = time;
         for (i = 0; i < n; i++) {
            snapshot_ptr->ticks[i] = ball[i].ticks;
            snapshot_ptr->dest[i] = ball[i].dest_hand;
         }
      } 
          /* check for ending condition.  Multiplexing and non-multi- */
	  /* plexing are separate cases to consider.                  */

      if (max_occupancy == 1) {
         if (!in_start && (pos == loop_pos) && (time > loop_time) &&
                        ((time % 2) == 0)) {	  
            print_time = do_time(print_time, time);
            if (l % 2) {
	       printf("swap hand\n");
               (void)do_time(0,0);
            }
	    printf("loop time %d\n", loop_time);
	    return;
	 }
      } else {            /* multiplexing case */
         new_ptr = snapshot_ptr;
         while (new_ptr != (struct snapshot *)0) {
            i = 1;
            if ((new_ptr->pos != pos) || (new_ptr->time == time))
               i = 0;
            if (i) {
               for (j = 0; (j < n) && i; j++)
                  if ((new_ptr->ticks[j] != ball[j].ticks) ||
                        (new_ptr->dest[j] != ball[j].dest_hand))
                     i = 0;
	       if (i) {
                  print_time = do_time(print_time, time);
	          printf("loop time %d\n", new_ptr->time);
	          return;
	       }

               for (j = 0, i = 1; (j < n) && i; j++)
                  if ((new_ptr->ticks[j] != ball[j].ticks) ||
                       (new_ptr->dest[j] == ball[j].dest_hand))
                     i = 0;
               if (i) {
                  print_time = do_time(print_time, time);
	          printf("swap hand\n");
                  (void)do_time(0, 0);     /* skip spaces */
                  printf("loop time %d\n", new_ptr->time);
	          return;
	       }
            }
            new_ptr = new_ptr->prev;
	 }
      }
      
          /* make any catches that need to be done.     */

      for (i = 0; i < n; i++) {
         if ((ball[i].ticks == 2) || (ball[i].ticks == -2)) {
            print_time = do_time(print_time, time);
	    if (max_occupancy == 1) {    /* in multiplexing mode? */
               if (ball[i].dest_hand == RIGHT)
                  printf("catch right\n");
               else
                  printf("catch left\n");
            } else {
               if (ball[i].dest_hand == RIGHT)
                  printf("catch right %s\n", colors[i]);
               else
                  printf("catch left %s\n", colors[i]);
            }
	 }
      }
      
          /* move the hands if we need to   */

      moveleft = moveright = 1;
      for (i = 0; i < n; i++) {
         if (ball[i].ticks == 1) {
	    if (ball[i].dest_hand == RIGHT) {
	       if (moveright) {
                  print_time = do_time(print_time, time);
	          printf("move right 0\n");
		  moveright = 0;
	       }
	    } else {
	       if (moveleft) {
                  print_time = do_time(print_time, time);
	          printf("move left 0\n");
		  moveleft = 0;
	       }
	    }
	 }
      }
      
          /* finally, make the throws that are required */

     if (!(time % 2)) {
      if (mode == ASYNCH) {
         for (i = 0; (i<max_occupancy)&&(j=throw[pos][0][i].value); i++) {
	    for (k = 0, q = 1; (k < n) && q; k++)
	       if (!ball[k].ticks && (ball[k].dest_hand == throwing_hand))
	          q = 0;
	    if (q)
	       die(SYNTAX);
	    ball[--k].dest_hand = ((j % 2) ? ((throwing_hand == RIGHT) ?
	                    LEFT : RIGHT) : throwing_hand);
	    if (j == 1) {
	       ball[k].ticks = -1;
               print_time = do_time(print_time, time);
	       if ((max_occupancy == 1) && !in_start) {
	          if (throwing_hand == RIGHT)
		     printf("throw right towards 4,1\n");
		  else
		     printf("throw left towards 4,1\n");
	       } else {
	          if (throwing_hand == RIGHT)
		     printf("throw right %s towards 4,1\n", colors[k]);
		  else
		     printf("throw left %s towards 4,1\n", colors[k]);
	       }
	    } else if (j != 2) {
	       ball[k].ticks = 2 * j;
               print_time = do_time(print_time, time);
	       if ((max_occupancy == 1) && !in_start) {
	          if (throwing_hand == RIGHT)
		     printf("throw right");
		  else
		     printf("throw left");
		  printf(" towards %d,%d\n",
		               ((j%2) ? 5 : -1), ball[k].ticks - 2);
	       } else {
	          if (throwing_hand == RIGHT)
		     printf("throw right");
		  else
		     printf("throw left");
		  printf(" %s towards %d,%d\n", colors[k],
		                    ((j%2) ? 5 : -1), ball[k].ticks - 2);
	       }
	    } else
	       ball[k].ticks = 2;    /* do a hold */
	 }
      } else {          /* synchronous mode */
         for (i = 0; i < hands; i++) {
	    for (j = 0; (j<max_occupancy)&&(k=throw[pos][i][j].value); j++) {
	       for (q = 0, r = 1; (q < n) && r; q++)
	          if (!ball[q].ticks && (ball[q].dest_hand == i))
		     r = 0;
	       if (r)
	          die(SYNTAX);
	       ball[--q].dest_hand = throw[pos][i][j].to;
	       ball[q].ticks = 2 * k;
	       if ((k != 2) || (ball[q].dest_hand != i)) {
                  print_time = do_time(print_time, time);
	          if ((max_occupancy == 1) && !in_start) {
		     if (i)
		        printf("throw left");
		     else
		        printf("throw right");
		     printf(" towards %d,%d\n", ((ball[q].dest_hand ==
		            i) ? -1 : 5), ball[q].ticks - 2);
		  } else {
		     if (i)
		        printf("throw left");
	             else
		        printf("throw right");
	             printf(" %s towards %d,%d\n", colors[q],
			    ((ball[q].dest_hand == i) ?
			    -1 : 5), ball[q].ticks - 2);
		  }
	       }
	    }
	 }
      }
     }
      
     if ((max_occupancy == 1) && (time == 2 * depth) && print_time)
                   /* loop label not printed yet */
        printf("%d:\n", time);
   
          /* decrement ball counters for next time around */

      for (i = 0; i < n; i++) {
         if (ball[i].ticks == -2)
	    ball[i].ticks = 0;
	 else if (ball[i].ticks != 0)
	    ball[i].ticks--;
      }

      if ((time % 2) == 0) {
         pos = (pos + 1) % l;
	 throwing_hand = ((throwing_hand == RIGHT) ? LEFT : RIGHT);
      }
      
      time++;          /* go to next time step */
   }

}


/*  Outputs in a format readable by Ed Carsten's JUGGLEPRO program  */

void print_jp(throw)
struct throw ***throw;
{
   int i, j, height, width, x, y, person;

   height = ((mode == PASSING) ? 4 : 2);

   if (mode == SYNCH)
      width = l;
   else
      width = ((l % 2) ? (2 * l) : l);

   printf("%dx%d\n", height, width);

   for (i = 0; i < height; i++) {
      for (j = 0; j < width; j++) {
         switch (mode) {
            case ASYNCH:
               if ((i + j) % 2)
                  x = y = 0;
               else {
                  y = throw[j % l][0][0].value;
                  x = ((y % 2) ? (i ? -1 : 1) : 0);
               }
               break;
            case SYNCH:     /* This isn't how Ed does it, but I can't get */
               if (j % 2)   /* the hands to work otherwise.               */
                  x = y = 0;
               else {
                  y = throw[j][i][0].value;
                  x = throw[j][i][0].to - i;
               }
               break;
            case PASSING:
               if ((i + j) % 2)
                  x = y = 0;
               else {
                  person = ((i > 1) ? 1 : 0);
                  y = throw[j % l][person][0].value;
                  x = ((y % 2) ? ((i % 2) ? -1 : 1) : 0);
                  if (throw[j % l][person][0].to != person)
                     x += 2;
                  x = ((x + i) % height) - i;
               }
               break;
         }

         printf("(%d,%d)", x, y);
      }
      printf("\n");
   }

   printf("%d\n", n);
}


void main(argc, argv)
int argc;
char **argv;
{
   int i, j, k, dest_format = JUGGLEPRO;
   char *pattern;
   struct throw ***throw;

   if (argc < 2) {
      printf("USAGE: j2anim <pattern string> [-options]\n\n");
      printf(
"Version 1.0 (Feb 92), by Jack Boyce (jboyce@tybalt.caltech.edu)\n\n");
      printf(
"This program takes as input a juggling pattern in the notation used by J2\n");
      printf(
"and converts it into forms readable by either S.Hawtin's JUG or Ed\n");
      printf(
"Carstens' JUGGLEPRO programs.  All modes of J2 except the custom mode are\n");
      printf(
"supported.  On most systems the input pattern should be given in quotes,\n");
      printf(
"since the characters `<([|])>' are interpreted as special symbols.\n\n");
      printf(
"Command line options:\n      -jug    Convert to JUG format\n");
      printf("      -jp     Convert to JUGGLEPRO format (default)\n\n");
      printf(
"Redirect the output to a file for loading into JUG or JUGGLEPRO.\n");
      printf("Example:   j2anim \"25[32]4[54]\" -jug >neat.jug\n");
      exit(0);
   }

   for (i = 2; i < argc; i++) {
      if (!strcmp(argv[i], "-jug"))
         dest_format = JUG;
      else if (!strcmp(argv[i], "-jp"))
         dest_format = JUGGLEPRO;
      else {
         printf("Unrecognized command line option\n");
         exit(0);
      }
   }

   pattern = argv[1];
   if (pattern[0] == '(') {        /* figure out the mode used */
      mode = SYNCH;
      hands = 2;
   } else if (pattern[0] == '<') {
      mode = PASSING;
      hands = 2;
   } else {
      mode = ASYNCH;
      hands = 1;
   }

   parse1(pattern, dest_format);     /* find n, l, and max_occupancy */

         /* Allocate the memory space that will hold the pattern. */
         /* The pattern will be stored internally in the same way */
         /* that J2 does.                                         */

   if (!(throw = (struct throw ***)malloc(l * sizeof(struct throw **))))
      die(MEMORY);
   for (i = 0; i < l; i++) {
      if (!(throw[i] = (struct throw **)malloc(hands * 
                                       sizeof(struct throw *))))
         die(MEMORY);
      for (j = 0; j < hands; j++) {
         if (!(throw[i][j] = (struct throw *)malloc(max_occupancy *
                                    sizeof(struct throw))))
            die(MEMORY);
         for (k = 0; k < max_occupancy; k++) {   /* clear throws */
            throw[i][j][k].value = 0;
            throw[i][j][k].to = j;
         }
      }
   }

   parse2(pattern, throw);    /* parse the pattern into memory */

   if (dest_format == JUG) {
      if ((mode != ASYNCH) && (mode != SYNCH))
         die(NO_JUG);
      print_jug(pattern, throw);
   } else {
      if (max_occupancy > 1)
         die(NO_JP);
      print_jp(throw);
   }

   exit(0);
}


void die(code)
int code;
{
   switch (code) {
      case SYNTAX:
         printf("Syntax error in input pattern\n");
         break;
      case MEMORY:
         printf("Insufficient memory\n");
         break;
      case TOO_MANY:
         printf("JUG can't do more than %d balls\n", NUM_BALLS);
         break;
      case NO_JUG:
         printf("JUG does not support passing\n");
         break;
      case NO_JP:
         printf("JUGGLEPRO does not support multiplexing\n");
         break;
   }
   exit(0);
}
